/*

* Copyright (c) 2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;

/**
 * Action operator in AQ plan.
 *
 * @see AQ
 */
public class ActionOperator extends Operator {
	
	/** The action that this action operator correponds to. */
	public Action action; 
	 
	/** Specification of the input parameters of the action. */
	public String actionParams [];
	
	boolean busy = false;  // indicate whether the action operator is busy with an action request or not
	
	Random ran = new Random ();
	
	GroupingThread gt;  // for multi-query optimization only
	boolean gt_start = false;
	
	CameraSimulator camSim [] = new CameraSimulator [CameraSimulator.MAX_REAL_CAMERA_NUM];
		
	String select_did;              // id of the selected device
	
	RealPhonePollingThread rppt_c;  // for current candidate phone
	
	String camera_id_c;             // for current candidate camera
	Point loc_c;
			
	
	/**
	 * Construct a new action operator.
	 *
	 * @param q the action-embedded query that this action operator belongs to
	 * @param action the action that this action operator correponds to. 
	 * @param actionParams the specification of the input parameters of the action
	 *
	 * @see Action
	 * @see AQ
	 */
	public ActionOperator(AQ q, Action action, String actionParams [])
	  {
	  	type = ACTION;
	  	this.q = q;
	  	this.action = action;
	  	this.actionParams = actionParams;
	  	
	  	for (int i=0;i<camSim.length;i++)
	  	  camSim[i] = null;
	  	  
	  	if (q.aortaqp.MULTI_QUERY_OPTIMIZATION && q.ACTION_OPTIMZATION)
	  	  gt = new GroupingThread(this);	  	  	  	
	  }
	  
	
	public void start() 
	  {
	  	if (q.aortaqp.MULTI_QUERY_OPTIMIZATION && q.ACTION_OPTIMZATION && !gt_start)
	  	  {
	  	  	gt.start();
	  	  	gt_start = true;
	  	  }
	  }
	  
	  
	public void stop() 
	  {
	  	if (q.aortaqp.MULTI_QUERY_OPTIMIZATION && q.ACTION_OPTIMZATION && gt_start)
	  	  {
	  	  	gt.stop();
	  	  	gt_start = false;
	  	  }
	  }
	  
	  
	public void getInput()
	  {
	  	if (q.type == Aorta.EVENT_DRIVEN_QUERY || q.type == Aorta.TRADITIONAL_CONTINUOUS_QUERY)
	  	  {
	  	  	int i;
	  	    long startTime, endTime;
	  	    Tuple t;
	  	    String actionParamsV [] = null;  // the array stores the string representation of the instantiated action parameter values on one selected candidate device
	  	
	  	    if (inputQueue.size() == 0) return;	  	    
	  	    
	  	    if (q.aortaqp.MULTI_QUERY_OPTIMIZATION && q.ACTION_OPTIMZATION)  // the grouping thread will handle the action workload scheduling
	  	      return;	  	        	      
	  	      
	  	    if ((!q.actionNesting && action == q.action) || (q.actionNesting && action == q.action2))
	  	      {
	  	  	    q.eventEndTime = System.currentTimeMillis();
	  	  	    q.eventStartTime= System.currentTimeMillis();
	    	    q.logFileWriter.println("Event reaction (Others) time: " + (q.eventEndTime - q.eventStartTime) + " ms.");
	            q.logFileWriter.println();
	            q.logFileWriter.flush();
	          }		  	
	  	    
	  	    action.q = q; 
	  	    
	  	    while (busy) {}
	   	    busy = true;   	     	        	    
	   	    
	   	    q.logFileWriter.println("Query " + q.name + " (id: " + q.id + ") executed once.");
	  	    q.logFileWriter.print("Action: " + action.name);
	  	    q.logFileWriter.print("(");
	  	    for (i=0;i<action.paramNum;i++)
	  	      if (i == action.paramNum - 1) q.logFileWriter.println(action.params[i].name + ")");
	  	      else q.logFileWriter.print(action.params[i].name + ", ");
	  	    
	  	    if ((q.selOp != null) && (q.selOp.outputOperator == this))
	  	      {
	  	      	q.selOp.getInput();
	  	      	
	  	      	if (q.actionNesting)
	  	      	  {
	  	      	  	t = (Tuple)inputQueue.elementAt(0);
	  	      	  	q.actionParams[1] = (String)t.fields.elementAt(0);	// image file name  	      	  	 
	  	      	  }	  	      	
	  	      	
	  	      	inputQueue.removeAllElements();  // remove the event tuple pushed from the scan operator
	  	      	
	  	      	if (inputQueue2.size() == 0)  // no candidate device
	  	      	  return;
	  	      	  	  	      	
	  	      	actionParamsV = selectCandidateDevice(inputQueue2);	  	      	
	  	      }
	  	    else
	  	      if ((q.joinOp != null) && (q.joinOp.outputOperator == this))
	  	        {
	  	          if (inputQueue.size() == 0)
	  	            return;
	  	            
	  	          actionParamsV = selectCandidateDevice(inputQueue);
	  	        }
	  	        
	  	    if (actionParamsV == null)
	  	      {
	  	      	q.logFileWriter.println("All candidate devices suitable for executing the action are currently unavailable. Fail to execute the action.");
	  	      	
	  	      	System.err.println("exec");
	   	        System.err.println("Query " + q.name + " (ID: " + q.id + ")  Action " + q.action.name);
	   	        System.err.println("All candidate devices suitable for executing the action are currently unavailable. Fail to execute the action");
	  	      	System.out.print("Aorta>");	  	      	
	   	      }
	  	    else  // execute the action on the selected candidate device
	  	      {
	  	      	startTime = System.currentTimeMillis();
	  	      	
	  	      	action.doAction(actionParamsV);
	  	      	
	  	      	if (q.actionNesting && this.outputOperator == q.actionOp) // second-level action
	  	      	  {
	  	      	  	Vector v = new Vector (5,5);
	  	      	  	v.addElement((String)((Photo)action).imageFileName);
	  	      	  	t = new Tuple(v);
	  	      	  	enQueue(t);	  	      	  	
	  	      	  }
	  	      	
	  	        // unlock the device
	  	        if (q.aortaqp.lockingEnabled)
	  	          {
	  	          	if (action.device.type.equals("sensor"))
	  	              q.aortaqp.lockSensorVector.removeElement((String)select_did);
	  	            else if (action.device.type.equals("camera"))
	  	                   q.aortaqp.lockCameraVector.removeElement((String)select_did);
	  	                 else if (action.device.type.equals("phone"))
	  	                        q.aortaqp.lockPhoneVector.removeElement((String)select_did);
	  	          }
	  	               
	  	      	endTime = System.currentTimeMillis();
	  	      	q.logFileWriter.println("Real action cost on the selected device: " + (endTime - startTime) + " ms.");
	  	      	q.logFileWriter.println("#");
	  	      	q.logFileWriter.println();
	  	      	q.logFileWriter.println();
	  	      	
	  	      	if (q.actionNesting && this.outputOperator == q.actionOp) // second-level action
	  	          outputOperator.getInput();
	  	      }	  	    
	  	   	  	    
	  	    q.logFileWriter.println();
	   	    q.logFileWriter.flush();	   	    
	   	     
	   	    busy = false;
	   	    
	   	    action.q = null;   	    
	  	  }	  	
	  }
	  
	
	
	/**
	 * Select the best one from a set of candidate devices to execute the action.
	 * Call corresponding methods based on the type of the candidate devices.
	 *
	 * @param v a vector each element of which contains information about one candidate device.
	 *
	 * @return the instantiated actio input parameter values that correspond to the selected device
	 *
	 * @see #selectCandidateSensor
	 * @see #selectCandidateCamera
	 * @see #selectCandidatePhone
	 */
	public String [] selectCandidateDevice (Vector v)
	  {
	  	if (action.device.type.equals("sensor"))
	  	  return selectCandidateSensor(v);
	  	else
	  	  if (action.device.type.equals("camera"))
	  	    return selectCandidateCamera(v);
	  	  else
	  	    if (action.device.type.equals("phone"))
	  	      return selectCandidatePhone(v);
	        else
	          return null;
	  }
	  
	public String [] selectCandidateSensor (Vector v)
	  {
	  	return null;  // not yet implemented
	  }
	 
    public String [] selectCandidateCamera (Vector v)
      {
      	int sel_no;
      	int num = v.size();
      	int i;	   	  
	   	long startTime, endTime;
	   	double minCost = Double.MAX_VALUE, estimatedCost;
	   	String camera_id;
	   	Point sensor_loc;	
	   	Tuple t; 
	   	CameraPollingThread cpt []; 
	   			   		  		   	
	   	
	   	startTime = System.currentTimeMillis();	   		   	
	   	if (q.ACTION_OPTIMZATION) // select the "best" candidate camera
	   	  {
	   	    q.logFileWriter.println("#");
	   	    q.logFileWriter.println("Polling the candidate cameras ... ");
	   	    q.logFileWriter.println();
	   	    
	   	    cpt = new CameraPollingThread[num];
	   	    for (i=0;i<num;i++)
	   	      {
	   	      	t = (Tuple)v.elementAt(i);
	   	      	camera_id = (String)t.fields.elementAt(0);	   	      	
	   	      	
	   	      	if (camSim[Integer.parseInt(camera_id)-1] == null)
	   	      	  camSim[Integer.parseInt(camera_id)-1] = new CameraSimulator(camera_id);
	   	      	  
	   	      	if (q.aortaqp.lockingEnabled)
	   	      	  if (q.aortaqp.lockCameraVector.contains(camera_id))
	   	            {cpt[i] = null; continue;} 
	   	           
	   	      	cpt[i] = new CameraPollingThread(camSim[Integer.parseInt(camera_id)-1], t, q.aortaqp.CAMERA_SIMULATION);
	   	      	cpt[i].start();
	   	      }
	   	    
	   	    endTime = System.currentTimeMillis();	   	    	        
	   	    while ( endTime - startTime < CameraSimulator.CONNECTION_TIMEOUT_LIMIT)
	   	      {
	   	        boolean finished = true;
	   	          	
	   	        for (i=0;i<num;i++)
	   	          if (cpt[i] != null && cpt[i].connected == false)
	   	          	finished = false;
	   	          	    	   	          	
	   	        if (finished) break;
	   	        
	   	        endTime = System.currentTimeMillis();	   	          	
	   	      }  	    
	   	      	    
	   	    minCost = Double.MAX_VALUE;
	   	    sel_no = -1;	   	    
	   	    for (i=0;i<num;i++)   // select the camera with the minimum action execution cost
	   	      {
	   	        t = (Tuple)v.elementAt(i);
	   	        camera_id_c = (String)t.fields.elementAt(0);  	        
	   	        loc_c = Point.parsePoint((String)t.fields.elementAt(4));
	   	        
	   	        if (cpt[i] == null) 
	   	          {
	   	          	q.logFileWriter.println("Candidate Camera " + camera_id_c + " is in use and has been locked.");
	   	           	continue;
	   	          }
	   	        
	   	        if (cpt[i].connected == false)
	   	          {
	   	          	 q.logFileWriter.println("Connection to candidate Camera " + camera_id_c + " times out");
      	             continue;
	   	          }
	   	        else
      	          q.logFileWriter.println("Connection cost to candidate Camera " + camera_id_c + ": " + camSim[Integer.parseInt(camera_id_c)-1].connectionCost + " ms.");      	
      		   	        	   	        
	   	        estimatedCost = estimateActionCost();	   	        	   	        
	   	        q.logFileWriter.println("Estimated action cost on candidate Camera " + camera_id_c + ": " + estimatedCost + " ms.");
	   	        q.logFileWriter.println();
	   	        if (minCost > estimatedCost)
	   	          {minCost = estimatedCost; sel_no = i;}  
	   	      }
	   	     
	   	    for (i=0;i<num;i++)
	   	      if (cpt[i] != null)
	   	        {
	   	          cpt[i].stop();
	   	          cpt[i] = null;
	   	        }	   	          	        
	   	        
	   	    System.gc();
	   	    
	   	    q.logFileWriter.println("Finish polling the candidate cameras.");
	   	    q.logFileWriter.println("#");
	   	    q.logFileWriter.println();
	   	    q.logFileWriter.flush();	   	    	   	          	   	          	
	   	  }
	    else // random selection
	   	  {
	   	  	if (num == 1) sel_no = 0;
	   	    else sel_no = Math.abs(ran.nextInt()) % num;
	   	    i = 0;
	   	    t = (Tuple)v.elementAt(sel_no);
	   	    camera_id = (String)t.fields.elementAt(0);	   	    
	   	    if (q.aortaqp.lockingEnabled)
	   	      {
	   	      	while (q.aortaqp.lockCameraVector.contains(camera_id))
	   	          {
	   	            q.logFileWriter.println("Candidate Camera " + camera_id + " is in use and has been locked.");
	   	            sel_no ++;
	   	            if (sel_no == num) sel_no = 0; 
	   	            i++; 
	   	            if (i == num) break;
	   	            t = (Tuple)v.elementAt(sel_no);
	   	            camera_id = (String)t.fields.elementAt(0);
	   	          }
	   	        if (q.aortaqp.lockCameraVector.contains(camera_id))
	   	          sel_no = -1;
	   	      }
	      }   
	   	
	   	if (sel_no == -1)
	   	  return null;   	  
	   	
        t = (Tuple)v.elementAt(sel_no);        
        camera_id = (String)t.fields.elementAt(0);
        sensor_loc = Point.parsePoint((String)t.fields.elementAt(4));
        camSim[Integer.parseInt(camera_id)-1].pan = camSim[Integer.parseInt(camera_id)-1].targetPan(sensor_loc);
        camSim[Integer.parseInt(camera_id)-1].tilt = camSim[Integer.parseInt(camera_id)-1].targetTilt(sensor_loc);
	   	camSim[Integer.parseInt(camera_id)-1].zoom = camSim[Integer.parseInt(camera_id)-1].targetZoom(sensor_loc);
	    
	    // lock the selected camera
	   	if (q.aortaqp.lockingEnabled)
	   	  q.aortaqp.lockPhoneVector.addElement(String.valueOf(camera_id));
	    select_did = camera_id;
	   	  
	   	// parse the values of the action input parameters of the selected camera
	   	String actionParamsV [] = new String [20];	   	
	   	int pos = 3;	   	
	   	for (i=0;i<action.paramNum;i++)
	      if (actionParams[i].charAt(1) == '.' && actionParams[i].charAt(0) != '.')
	   	    actionParamsV[i] = (String)t.fields.elementAt(pos++);
	   	  else
	   	    actionParamsV[i] = actionParams[i];	
	   	    
	   	actionParamsV[i++]= (String)t.fields.elementAt(0);  // id
	   	actionParamsV[i++] = (String)t.fields.elementAt(1); // username
	   	actionParamsV[i] = (String)t.fields.elementAt(2);   // password  
	   	 
	    q.logFileWriter.println("#");	   	  
	    q.logFileWriter.println("Camera " + t.fields.elementAt(0) + " is selected to execute the action.");
	    q.logFileWriter.print("Instantiated action input parameter values: ");
	    for (i=0;i<action.paramNum;i++)
	   	  q.logFileWriter.print(action.params[i].name + " = " + actionParamsV[i] + " ");
	    q.logFileWriter.println();
	    endTime = System.currentTimeMillis();
	    q.logFileWriter.println("Optimization Time: " + (endTime - startTime) + " ms.");
	    if (q.ACTION_OPTIMZATION)
	      q.logFileWriter.println("Estimated action cost on the selected camera: " + minCost + " ms.");	   	  
	    q.logFileWriter.flush();
	    	   	  	   	  
	    v.removeAllElements();      
        
        return actionParamsV;
      }      
   
      
    public String [] selectCandidatePhone (Vector v)
      {
      	  int num = v.size();  // number of candidate phones
	   	  int sel_no;
	   	  int i;
	   	  long startTime, endTime;
	   	  double minCost = Double.MAX_VALUE, estimatedCost;
	   	  String phone_id, phone_no;
	   	  Tuple t;
	   	  RealPhonePollingThread rppt [];
	   	  	   	  
	   	  startTime = System.currentTimeMillis();	   	  
	   	  if (q.aortaqp.PHONE_EMULATION == true)  // we are using J2ME phone emulator
	   	    {
	   	      if (num == 1) sel_no = 0;
	   	      else sel_no = (int)(Math.abs(ran.nextInt()) % num);  // random selection
	   	      
	   	      i=0;
	   	      String str;
	   	      if (q.aortaqp.lockingEnabled)
	   	        {
	   	          while(q.aortaqp.lockPhoneVector.contains(str = (String)((Tuple)v.elementAt(sel_no)).fields.elementAt(0)))
	   	            {
	   	              q.logFileWriter.println("Candidate Phone " + str + " is in use and has been locked.");
	   	              sel_no ++;
	   	              if (sel_no == num) sel_no = 0; 
	   	              i++; 
	   	              if (i == num) break;
	   	            }
	   	          if (q.aortaqp.lockPhoneVector.contains((String)((Tuple)v.elementAt(sel_no)).fields.elementAt(0)))
	   	            sel_no = -1;  // all candidate phones has been locked
	   	        }   	                	        
	   	    }	   	    
	   	  else  // we are using real cell phones 	  
	   	    {
	   	       if (q.ACTION_OPTIMZATION)  // select the "best" candidate phone
	   	          {
	   	          	q.logFileWriter.println("#");
	   	            q.logFileWriter.println("Polling the candidate phones ... ");
	   	          	q.logFileWriter.println();
	   	          	
	   	          	rppt = new RealPhonePollingThread [num];
	   	            for (i=0;i<num;i++)
	   	    	      {
	   	    	  	    t = (Tuple)v.elementAt(i);	   	    	  	
	   	                phone_id = (String)t.fields.elementAt(0); // id of the phone
	   	    	  	    phone_no = (String)t.fields.elementAt(1);  // no of the phone
	   	    	  	
	   	    	     	// the candidate phone has been locked
	   	    	  	    if (q.aortaqp.lockingEnabled)
	   	    	  	      if (q.aortaqp.lockPhoneVector.contains(phone_id))
	   	    	  	        {rppt[i] = null; continue;}	   	    	  	
	   	    	  		
	   	    	  	    rppt[i] = new RealPhonePollingThread(this, RealPhonePollingThread.INITIAL_PORT_NO+Integer.parseInt(phone_id), phone_no);
	   	    	  	    rppt[i].start();	   	    	  	
	   	    	      } 
	   	    	   	   	    	  
	   	            endTime = System.currentTimeMillis();	   	        
	   	            while ( endTime - startTime < RealPhonePollingThread.PHONE_POLLING_TIMEOUT)
	   	              {
	   	          	    boolean finished = true;
	   	          	
	   	          	    for (i=0;i<num;i++)
	   	          	      if (rppt[i] != null && rppt[i].connected == false)
	   	          	        finished = false;
	   	          	    	   	          	
	   	          	    if (finished) break;
	   	          	    
	   	          	    endTime = System.currentTimeMillis();	   	          	
	   	              } 
	   	          
	   	            sel_no = -1;  	          	
	   	          	minCost = Double.MAX_VALUE;	   	       
	   	            for (i=0;i<num;i++)   // select the phone with the minimum action execution cost
	   	              {
	   	          	    t = (Tuple)v.elementAt(i);
	   	          	    if (rppt[i] == null) 
	   	          	      {
	   	          	      	q.logFileWriter.println("Candidate Phone " + t.fields.elementAt(0) + " is in use and has been locked.");
	   	          	      	continue;
	   	          	      }
	   	          	    if (rppt[i].connected == false)
	   	          	      {
	   	          	      	q.logFileWriter.println("Connection to candidate Phone " + t.fields.elementAt(0) + " times out");
      	          	      	continue;
	   	          	      }
	   	          	    else
      	                  q.logFileWriter.print("Connection cost to candidate Phone " + t.fields.elementAt(0) + ": " + rppt[i].responseTime + " ms.");      	
      	
	   	          	    rppt_c = rppt[i];
	   	          	    estimatedCost = estimateActionCost();
	   	          	    q.logFileWriter.print("Estimated action cost on candidate Phone " + t.fields.elementAt(0) + ": " + estimatedCost + " ms.");      	
      	                q.logFileWriter.println();
	   	          	    if (minCost > estimatedCost)
	   	          	      {minCost = estimatedCost; sel_no = i;}  
	   	          	  }
	   	          	
	   	          	for (i=0;i<num;i++)
	   	              if (rppt[i] != null)
	   	                {
	   	                  rppt[i].stop();
	   	                  rppt[i] = null;
	   	                }	   	                  	        
	   	        
	   	            System.gc();
	   	            
	   	            q.logFileWriter.println("Finish polling the candidate phones.");
	   	            q.logFileWriter.println("#");
	   	            q.logFileWriter.println();
	   	            q.logFileWriter.flush();	   	          	   	          	
	   	          }
	   	        else  // random selection
	   	          {
	   	          	if (num == 1) sel_no = 0;
	   	          	else sel_no = Math.abs(ran.nextInt()) % num;
	   	          	i = 0;
	   	          	t = (Tuple)v.elementAt(sel_no);	   	    	  	
	   	            phone_id = (String)t.fields.elementAt(0);
	   	          	if (q.aortaqp.lockingEnabled)
	   	          	  {
	   	          	  	while (q.aortaqp.lockPhoneVector.contains(phone_id))
	   	          	      {
	   	          	  	    q.logFileWriter.println("Candidate Phone " + phone_id + " is in use and has been locked.");
	   	          	  	    sel_no ++;
	   	          	  	    if (sel_no == num) 
	   	          	  	      sel_no = 0; 
	   	          	  	    i++; 
	   	          	  	    if (i == num) break;
	   	          	  	    t = (Tuple)v.elementAt(sel_no);	   	    	  	
	   	                    phone_id = (String)t.fields.elementAt(0);
	   	          	      }
	   	          	    if (q.aortaqp.lockPhoneVector.contains(phone_id))
	   	          	      sel_no = -1;
	   	          	  }
	   	          }   	            
	   	        	    	
	   	    }   	  
	   	  
	   	  if (sel_no == -1) 
	   	    return null;
	   	       
	   	  t = (Tuple)v.elementAt(sel_no);
	   	  phone_id = (String)t.fields.elementAt(0);
	   	  
	   	  // lock the selected phone
	   	  if (q.aortaqp.lockingEnabled)
	   	    q.aortaqp.lockPhoneVector.addElement(String.valueOf(phone_id));
	   	  select_did = phone_id;
	   	  
	   	  // parse the values of the action input parameters of the selected phone
	   	  String actionParamsV [] = new String [20]; 
	   	  int pos = 1;
	   	  for (i=0;i<action.paramNum;i++)
	   	    if (actionParams[i].charAt(1) == '.' && actionParams[i].charAt(0) != '.')
	   	      actionParamsV[i] = (String)t.fields.elementAt(pos++);
	   	    else
	   	      actionParamsV[i] = actionParams[i];	  
	   	 
	   	  q.logFileWriter.println("#");	   	  
	   	  q.logFileWriter.println("Phone " + t.fields.elementAt(0) + " is selected to execute the action.");
	   	  q.logFileWriter.print("Instantiated action input parameter values: ");
	   	  for (i=0;i<action.paramNum;i++)
	   	    q.logFileWriter.print(action.params[i].name + " = " + actionParamsV[i] + " ");
	   	  q.logFileWriter.println();
	   	  endTime = System.currentTimeMillis();
	   	  q.logFileWriter.println("Optimization Time: " + (endTime - startTime) + " ms.");
	   	  if (q.ACTION_OPTIMZATION && !q.aortaqp.PHONE_EMULATION)
	   	    q.logFileWriter.println("Estimated action cost on the selected phone: " + minCost + " ms.");	   	  
	   	  q.logFileWriter.flush();
	   	  	   	  	   	  
	   	  v.removeAllElements();	   	  
	   	  
	   	  return actionParamsV;
      }
      
    
    /**
     * Estimate the cost of an action execution on a candidate device. 
     *
     * @return the estimated cost of the action on the device
     *
     * @see Action
     * @see #estimateOperationSequenceCost 
     **/
    public double estimateActionCost ()
      {
      	return estimateOperationSequenceCost(action.opSeq);   	          	      
      }
      
   
   /** 
    * Estimated the cost of an operation sequence on a candidate device.
    *
    * @param opSeq an instance of {@link OperationSequence}
    *
    * @return the estimated cost of the operation sequence on the device
    *
    * @see OperationSequence
    * @see #estimateOperationSetCost
    * @see #estimateOperationCost
    */
   public double estimateOperationSequenceCost(OperationSequence opSeq)
     {
     	int i;
     	double estimatedCost = 0;
     	
     	for (i=0;i<opSeq.opUnitNum;i++)
     	  if (opSeq.op_units[i].flag == OperationUnit.OPERATION_SEQUENCE)
     	    estimatedCost += estimateOperationSequenceCost((OperationSequence)opSeq.op_units[i]);
     	  else
     	    if (opSeq.op_units[i].flag == OperationUnit.OPERATION_SET)
     	      estimatedCost += estimateOperationSetCost((OperationSet)opSeq.op_units[i]);
     	    else
     	      if (opSeq.op_units[i].flag == OperationUnit.OPERATION)
     	      estimatedCost += estimateOperationCost((Operation)opSeq.op_units[i]);
     	        	
     	return estimatedCost;
     }
     
   
   /** 
    * Estimated the cost of an operation set on a candidate device.
    *
    * @param opSet an instance of {@link OperationSet}
    *
    * @return the estimated cost of the operation set on the device
    *
    * @see OperationSet
    * @see #estimateOperationSequenceCost
    * @see #estimateOperationCost
    */
   public double estimateOperationSetCost(OperationSet opSet)
     {
     	int i;
     	double estimatedCost = 0;     	
     	
     	for (i=0;i<opSet.opUnitNum;i++)
     	  if (opSet.op_units[i].flag == OperationUnit.OPERATION_SEQUENCE)
     	    estimatedCost = Math.max(estimatedCost,estimateOperationSequenceCost((OperationSequence)opSet.op_units[i]));
     	  else
     	    if (opSet.op_units[i].flag == OperationUnit.OPERATION_SET)
     	      estimatedCost = Math.max(estimatedCost, estimateOperationSetCost((OperationSet)opSet.op_units[i]));
     	    else
     	      if (opSet.op_units[i].flag == OperationUnit.OPERATION)
     	        estimatedCost = Math.max(estimatedCost,estimateOperationCost((Operation)opSet.op_units[i]));
     	        	
     	return estimatedCost;
     }
     
     
   /** 
    * Estimated the cost of an operation on a candidate device.
    * Call corresponding methods based on the type of the candidate devices.
    *
    * @param op an instance of {@link Operation}
    *
    * @return the estimated cost of the operation on the device
    *
    * @see Operation
    * @see #estimateSensorOperationCost
    * @see #estimateCameraOperationCost
    * @see #estimatePhoneOperationCost
    */
   public double estimateOperationCost(Operation op)
     {
     	if (action.device.type.equals("sensor"))
     	  return estimateSensorOperationCost(op);
     	else
     	  if (action.device.type.equals("camera"))
     	    return estimateCameraOperationCost(op);
     	  else
     	    if (action.device.type.equals("phone"))
     	      return estimatePhoneOperationCost(op);
     	    else
     	      return Double.MAX_VALUE;    	      	     	
     }
     
   public double estimateSensorOperationCost(Operation op)
     {
       return Double.MAX_VALUE;  // not yet implemented	
     }
     
   public double estimateCameraOperationCost(Operation op)
     {
     	int number;
     	int id = Integer.parseInt(camera_id_c);
     	int start,stop;
     	int i;
     	String str, paramSymbol, paramAttribute;
     	
     	if (op.atomOp.name.equals("connect"))
     	  {
     	  	try {number = Integer.parseInt(op.number);}
     	  	catch (Exception e) {return Double.MAX_VALUE;}     	  		  	
     	  	return camSim[id-1].connectionCost*number;
          }
          
        if (op.atomOp.name.equals("takeSmallSizePhoto") || op.atomOp.name.equals("takeMediumSizePhoto") || op.atomOp.name.equals("takeLargeSizePhoto") )
     	  {
     	  	try {number = Integer.parseInt(op.number);}
     	  	catch (Exception e) {return Double.MAX_VALUE;}     	  	   	  	
     	  	return op.atomOp.estimatedCost*number;
     	  }
     
        if (op.atomOp.name.equals("pan"))
          {
          	str = op.number;  // either a digital number, or must be the system-provided location function deltaPan(current_pan, target_location)
            
            start = str.indexOf("deltaPan(");
            if (start == -1)  
              {
              	try {number = Integer.parseInt(op.number);}
              	catch (Exception e) {return Double.MAX_VALUE;}  // unknown number specification
              	return number * op.atomOp.estimatedCost;
              }
            start += "deltaPan(".length();
            stop = str.indexOf(",",start);
            if (stop == -1) 
              return Double.MAX_VALUE;
            paramSymbol = str.substring(start,stop).trim();
            paramAttribute = Aorta.NULL_STRING;
            
            for (i=0;i<action.device.dpsNum;i++)
              if (action.device.dps[i].symbol.equals(paramSymbol))
                {paramAttribute = action.device.dps[i].attribute; break;}            
            if (!paramAttribute.equals("pan"))
              return Double.MAX_VALUE;
            start = stop + 1;
            stop = str.indexOf(")",start);
            if (stop == -1)
              return Double.MAX_VALUE;
            paramSymbol = str.substring(start,stop).trim();
            for (i=0;i<action.paramNum;i++)
              if (action.params[i].symbol.equals(paramSymbol)) 
                {paramAttribute = action.params[i].name; break;}
            if (!paramAttribute.equals("location"))
              return Double.MAX_VALUE;
            return op.atomOp.estimatedCost * camSim[id-1].deltaPan(camSim[id-1].pan, loc_c);         	
          }
          
        if (op.atomOp.name.equals("tilt"))
          {
          	str = op.number;
            
            start = str.indexOf("deltaTilt(");
            if (start == -1)  
              {
              	try {number = Integer.parseInt(op.number);}
              	catch (Exception e) {return Double.MAX_VALUE;}  // unknown number specification
              	return number * op.atomOp.estimatedCost;
              }
            start += "deltaTilt(".length();
            stop = str.indexOf(",",start);
            if (stop == -1) 
              return Double.MAX_VALUE;
            paramSymbol = str.substring(start,stop).trim();
            paramAttribute = Aorta.NULL_STRING;
            
            for (i=0;i<action.device.dpsNum;i++)
              if (action.device.dps[i].symbol.equals(paramSymbol))
                {paramAttribute = action.device.dps[i].attribute; break;}
            if (!paramAttribute.equals("tilt"))
              return Double.MAX_VALUE;
            start = stop + 1;
            stop = str.indexOf(")",start);
            if (stop == -1)
              return Double.MAX_VALUE;
            paramSymbol = str.substring(start,stop).trim();
            for (i=0;i<action.paramNum;i++)
              if (action.params[i].symbol.equals(paramSymbol)) 
                {paramAttribute = action.params[i].name; break;}
            if (!paramAttribute.equals("location"))
              return Double.MAX_VALUE;                  
            return op.atomOp.estimatedCost * camSim[id-1].deltaTilt(camSim[id-1].tilt, loc_c);
          }
          
        if (op.atomOp.name.equals("zoom"))
          {
          	str = op.number;
            
            start = str.indexOf("deltaZoom(");
            if (start == -1)  
              {
              	try {number = Integer.parseInt(op.number);}
              	catch (Exception e) {return Double.MAX_VALUE;}  // unknown number specification
              	return number * op.atomOp.estimatedCost;
              }
            start += "deltaZoom(".length();
            stop = str.indexOf(",",start);
            if (stop == -1) 
              return Double.MAX_VALUE;
            paramSymbol = str.substring(start,stop).trim();
            paramAttribute = Aorta.NULL_STRING;
            
            for (i=0;i<action.device.dpsNum;i++)
              if (action.device.dps[i].symbol.equals(paramSymbol))
                {paramAttribute = action.device.dps[i].attribute; break;}
            if (!paramAttribute.equals("zoom"))
              return Double.MAX_VALUE;
            start = stop + 1;
            stop = str.indexOf(")",start);
            if (stop == -1)
              return Double.MAX_VALUE;
            paramSymbol = str.substring(start,stop).trim();
            for (i=0;i<action.paramNum;i++)
              if (action.params[i].symbol.equals(paramSymbol)) 
                {paramAttribute = action.params[i].name; break;}
            if (!paramAttribute.equals("location"))
              return Double.MAX_VALUE;              
            return op.atomOp.estimatedCost * camSim[id-1].deltaZoom(camSim[id-1].zoom, loc_c);
          }
          
     	// unknown atomic operation
     	return Double.MAX_VALUE;  	
     }
     
   public double estimatePhoneOperationCost(Operation op)
     {
     	int number;
     	int start, stop;
     	String str, paramSymbol, paramAttribute;
     	
     	if (op.atomOp.name.equals("connect"))
     	  {
     	  	try {number = Integer.parseInt(op.number);}
     	  	catch (Exception e) {return Double.MAX_VALUE;}
     	  	return rppt_c.responseTime*number;
          }
     	
     	if (op.atomOp.name.equals("receiveTextMessage"))
     	  {
     	  	str = op.number;
            
            start = str.indexOf("isEnoughMem(");
            if (start == -1)  // unknown number specification
              return Double.MAX_VALUE;
            start += "isEnoughMem(".length();
            stop = str.indexOf(",",start);
            if (stop == -1) 
              return Double.MAX_VALUE;
            try {number = Integer.parseInt(str.substring(start,stop).trim());}
            catch (Exception e) {return Double.MAX_VALUE;}
            if (number <=0) 
              return Double.MAX_VALUE;
            start = stop + 1;
            stop = str.indexOf(")",start);
            if (stop == -1)
              return Double.MAX_VALUE;
            paramSymbol = str.substring(start,stop).trim();
            paramAttribute = Aorta.NULL_STRING;
            for (int i=0;i<action.device.dpsNum;i++)
              if (action.device.dps[i].symbol.equals(paramSymbol))
                {paramAttribute = action.device.dps[i].attribute; break;}            
            if (!paramAttribute.equals("freeMem"))
              return Double.MAX_VALUE;            
            if (rppt_c.freeMem < number*SendMsg.MESSAGE_SIZE_LIMIT)
              return Double.MAX_VALUE;
            else
              return number*op.atomOp.estimatedCost;              	  	  
     	  }
     	
     	if (op.atomOp.name.equals("receivePhoto"))
     	  {
     	  	str = op.number;
            
            start = str.indexOf("isEnoughMem(");
            if (start == -1)  // unknown number specification
              return Double.MAX_VALUE;
            start += "isEnoughMem(".length();
            stop = str.indexOf(",",start);
            if (stop == -1) 
              return Double.MAX_VALUE;
            try {number = Integer.parseInt(str.substring(start,stop).trim());}
            catch (Exception e) {return Double.MAX_VALUE;}
            if (number <=0) 
              return Double.MAX_VALUE;
            start = stop + 1;
            stop = str.indexOf(")",start);
            if (stop == -1)
              return Double.MAX_VALUE;
            paramSymbol = str.substring(start,stop).trim();
            paramAttribute = Aorta.NULL_STRING;
            for (int i=0;i<action.device.dpsNum;i++)
              if (action.device.dps[i].symbol.equals(paramSymbol))
                {paramAttribute = action.device.dps[i].attribute; break;}            
            if (!paramAttribute.equals("freeMem"))
              return Double.MAX_VALUE;            
            if (rppt_c.freeMem < number*SendPhoto.PHOTO_SIZE_LIMIT)
              return Double.MAX_VALUE;
            else
              return number*op.atomOp.estimatedCost;
     	  }
     	  
     	// unknown atomic operation
     	return Double.MAX_VALUE;
     }	  
}