/*

* Copyright (c) 2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;

/**
 * A simulator that simulates the action workload scheduling process   
 * within the shared photo() action operator in Aorta.                
 *                                                         
 * It is developed for the sake of convenient performance evaluation  
 * and comparison of different scheduling algorithms. 
 */ 
public class AortaSchedulingSimulator {
	
	public static final byte UNIFORM = 0x01;
	public static final byte SKEWED = 0x02;
	byte distribution;
	
	int cameraNum;
	int requestNum;
	double skewness;
	
	
	/**
	 * @param distribution the scheduling workload distribution, either {@link #UNIFORM} or {@link #SKEWED}
	 * @param cameraNum the number of cameras involved in the scheduling
	 * @param requestNum the number of action requets contained in the workload
	 * @param skewness for skewed workload only, see {@link SkewedSyntheticWorkload#skewness}
	 */
	public AortaSchedulingSimulator (byte distribution, int cameraNum, int requestNum, double skewness)
	  {
	  	this.distribution = distribution;
	  	this.cameraNum = cameraNum;
	  	this.requestNum = requestNum;
	  	this.skewness = skewness;
	  }
	
	
	/**
	 * Generate a synthetic workload with uniform or skewed workload distribution and then
	 * perform the five scheduling algorithms we study on the workload generated. 
	 *
	 * @see ActionRequest
	 * @see UniformSyntheticWorkload
	 * @see SkewedSyntheticWorkload
	 * @see LERFA
	 * @see SRFE
	 * @see SRFAE
	 * @see SA
	 * @see LS
	 * @see RANDOM
	 */
	public void schedule ()
	  {	
		long startTime, endTime;
		String logFile;
		Vector requests = null, requests_backup = null;
		PrintWriter logFileWriter = null;
		
 					
		System.out.println("\t\t     Aorta Action Workload Scheduling Simulator");
		System.out.println("A grouping time interval begins.");
		  
		if (distribution == UNIFORM)
		  {
		  	UniformSyntheticWorkload usw = new UniformSyntheticWorkload(cameraNum, requestNum);
	        requests = usw.generate();
	     }
	    else if (distribution == SKEWED)
	           {
	      	     SkewedSyntheticWorkload ssw = new SkewedSyntheticWorkload(cameraNum, requestNum, skewness);
	             requests = ssw.generate();	
	           }
	         else
	           {
	           	 System.err.println("Invalid workload distribution specification.");
	           	 System.exit(1);
	           }
	                       
	    // LERFA + SRFE
		System.out.println("LERFA + SRFE ... ");
		logFile = "../logs/scheduling_logs/LERFA+SRFE.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		 	  				
		requests_backup = new Vector(30,10);
		backupRequests(requests, requests_backup);  // backup the workload
		startTime = System.currentTimeMillis();
	    LERFA lerfa = new LERFA (requests);
		lerfa.assignment();
		endTime = System.currentTimeMillis();
		logFileWriter.println();		 
		logFileWriter.println("LERFA assignment time: " + (endTime - startTime) + " ms.");
					    	  			   	  
		SRFE srfe = new SRFE (requests, logFileWriter, endTime - startTime);
		srfe.execution();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();			   	  
			  
		 
		/* // LERFA + LRFE
		System.out.println("LERFA + LRFE ... ");
		logFile = "../logs/scheduling_logs/LERFA+LRFE.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		 	  				
		requests_backup = new Vector(30,10);
		backupRequests(requests, requests_backup);
		startTime = System.currentTimeMillis();
	    lerfa = new LERFA (requests);
		lerfa.assignment();
		endTime = System.currentTimeMillis();
		logFileWriter.println();		 
		logFileWriter.println("LERFA assignment time: " + (endTime - startTime) + " ms.");
					    	  			   	  
		LRFE lrfe = new LRFE (requests, logFileWriter, endTime - startTime);
		lrfe.execution();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();			   	  
			  		   	  
		
		// LERFA + RRE
		System.out.println("LERFA + RRE ... ");
		logFile = "../logs/scheduling_logs/LERFA+RRE.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		 	  				
		requests_backup = new Vector(30,10);
		backupRequests(requests, requests_backup);
		startTime = System.currentTimeMillis();
	    lerfa = new LERFA (requests);
		lerfa.assignment();
		endTime = System.currentTimeMillis();
		logFileWriter.println();		 
		logFileWriter.println("LERFA assignment time: " + (endTime - startTime) + " ms.");
					    	  			   	  
		RRE rre = new RRE (requests, logFileWriter, endTime - startTime);
		rre.execution();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close(); */
		   	  
			  		   	  
		// SRFAE
		System.out.println("SRFAE ... ");
		logFile = "../logs/scheduling_logs/SRFAE.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		logFileWriter.println();
 	  				
		requests_backup = new Vector(30,10);
		backupRequests(requests, requests_backup);
		SRFAE srfae = new SRFAE (requests, logFileWriter);
		srfae.execution();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();			   	  
					 
			   
		/* // LRFAE 
		System.out.println("LRFAE ... ");
		logFile = "../logs/scheduling_logs/LRFAE.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		logFileWriter.println();
 	  				
		requests_backup = new Vector(30,10);
		backupRequests(requests, requests_backup);
		LRFAE lrfae = new LRFAE (requests, logFileWriter);
		lrfae.execution();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close(); */
			
	
		// RANDOM
	    System.out.println("RANDOM ... ");
		logFile = "../logs/scheduling_logs/RANDOM.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		 	  				
 	  	requests_backup = new Vector(30,10);
		backupRequests(requests, requests_backup);
		RANDOM random = new RANDOM (requests, logFileWriter);
		random.execution();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();		
				  
		// LS
		System.out.println("LS ... ");
		logFile = "../logs/scheduling_logs/LS.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		logFileWriter.println();
		 	  				
 	  	requests_backup = new Vector(30,10);
		backupRequests(requests, requests_backup);
		LS ls = new LS (requests, logFileWriter);
		ls.doAlgorithm();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();		
				  
		
		/* // LRFLS
		System.out.println("LRFLS ... ");
		logFile = "../logs/scheduling_logs/LRFLS.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		logFileWriter.println();
		 	  				
 	  	requests_backup = new Vector(30,10);
		backupRequests(requests, requests_backup);
		LRFLS lrfls = new LRFLS (requests, logFileWriter);
		lrfls.doAlgorithm();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();		
		
			    
		// SRFLS
		System.out.println("SRFLS ... ");
		logFile = "../logs/scheduling_logs/SRFLS.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		logFileWriter.println();
		 	  				
 	  	requests_backup = new Vector(30,10);
		backupRequests(requests, requests_backup);
		SRFLS srfls = new SRFLS (requests, logFileWriter);
		srfls.doAlgorithm();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close(); */
				   	
			
		// SA
		System.out.println("SA ... ");
		logFile = "../logs/scheduling_logs/SA.log";
		try {logFileWriter = new PrintWriter (new FileOutputStream (logFile));}
		catch (Exception e) {e.printStackTrace();}
		logFileWriter.println("A grouping time interval begins.");
		logFileWriter.println();
		 	  				
 	  	requests_backup = new Vector(30,10);
		backupRequests(requests, requests_backup);
		SA sa = new SA (requests, logFileWriter);
		sa.execution();
		requests = requests_backup;
		
		logFileWriter.println();
		logFileWriter.println("The grouping time interval ends.");
		logFileWriter.flush();
		logFileWriter.close();
			   
		
        System.out.println("The grouping time interval ends.");       
      }  
      
      
    /** 
     * Backup the generated synthetic workload so that the same workload can be used for all five scheduling algorithm.
     * 
     * @param requests the scheduling workload for backup
     * @param requests_backup a backup of the workload
     */
    public void backupRequests (Vector requests, Vector requests_backup)
      {
      	int i;
      	for (i=0;i<requests.size();i++)
		  requests_backup.addElement((ActionRequest)requests.elementAt(i));
      }
	
	
	/** Specify the parameters of the scheduling simulator and start the simulator. */
	public static void main (String argv [])  
	  {
	  	// change these parameters in different experiments
	  	byte distribution = AortaSchedulingSimulator.UNIFORM;
	  	
	  	int cameraNum = 10;
	  	
	  	int requestNum = 20;
	  	
	  	double skewness = 1;
	  	
	  	AortaSchedulingSimulator aortass = new AortaSchedulingSimulator(distribution,cameraNum,requestNum, skewness);
	  	
	  	aortass.schedule();
	  }	
} 