/*

* Copyright (c) 2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.io.*;
import java.util.*;

/**
 * Implementation of Algorithm 2 (SRFAE: Shorest Request First Assignment and Execution).
 * 
 * @see LRFAE
 * @see AortaSchedulingSimulator#schedule
 * @see GroupingThread#scheduleCameras
 */ 
public class SRFAE {
 	
 	Vector requests;
	
	PrintWriter logFileWriter;
	
	Vector didVector = new Vector (20,10);
	
	CameraSimulator camSim [];
	double singleServiceTime [];
	double overallServiceTime, overallSchedulingTime;
	
	ActionRequest sel_ar;
	String sel_did;
	Tuple sel_status;
	double total_actionCost, sel_actionCost;	
	
		
 	/** 
	 * @param requests a vector of action requests to be scheduled
	 * @param logFileWriter A {@link PrintWriter} corresponds to the log file
	 * 
	 * @see ActionRequest
	 */
	 public SRFAE (Vector requests, PrintWriter logFileWriter) 
 	  {
 		this.requests = requests;
		this.logFileWriter = logFileWriter;	
		
		overallServiceTime = 0;  overallSchedulingTime = 0; 
		
		long startTime = System.currentTimeMillis(); 				
 		 			
 		initialize();
 		
 		long endTime = System.currentTimeMillis();
 		
 		overallSchedulingTime += endTime - startTime;		  
 	  } 
 	
 	
 	public void initialize ()
 	  {
 	  	// decide the set of candidate devices
		int i,j;
    	for (i=0;i<requests.size();i++)
    	  for (j=0;j<((ActionRequest)requests.elementAt(i)).didVector.size();j++)
    	  	if (!didVector.contains((String)((ActionRequest)requests.elementAt(i)).didVector.elementAt(j)))
    	  	  didVector.addElement(((String)((ActionRequest)requests.elementAt(i)).didVector.elementAt(j)));
    	
    	camSim = new CameraSimulator [didVector.size()];
    	singleServiceTime = new double [didVector.size()];    	
       	for (i=0;i<didVector.size();i++)
    	  {
    	  	camSim[i] = new CameraSimulator(String.valueOf(i+1));
    	  	singleServiceTime[i] = 0;    	  	
    	  }  	  	
 	  }
 	  
 	  
 	/** Assign and execute the requests on the candidate cameras with the shorest request first heurisic. */
 	public void execution ()  
 	  {
 		int i;
 		int count = 0;
 		int id;
 		long startTime, endTime;
 		double connectionCost, actionCost; 	
 		
 		connectionCost = 0;
 		for (i=0;i<camSim.length;i++)
 		  if (connectionCost < camSim[i].connection())
 		    connectionCost = camSim[i].connectionCost; 
 		overallSchedulingTime += connectionCost;	
 		
 		while (requests.size() > 0)
 		  {
 		  	logFileWriter.println("#"); 		  	
             		  	
 		  	startTime = System.currentTimeMillis(); 		  	
 		  	shorestRequestSelection();
 		  	endTime = System.currentTimeMillis();
 		  	overallSchedulingTime += endTime - startTime;
 		  	  	
 		  	id = Integer.parseInt(sel_did);
 		  	sel_actionCost = total_actionCost - singleServiceTime[id-1];
 		  	singleServiceTime[id-1] = total_actionCost;  		  	 		  	
 		  	
 		  	logFileWriter.println("Round #" + (++count) + ": Request #" + sel_ar.aid +  " is serviced on Camera " + sel_did + ".");
 		  	logFileWriter.println("Action cost: " + sel_actionCost + " ms.");
 		  	logFileWriter.println("#");
 		  	logFileWriter.println();
 		  	 		  	
 		  	camSim[id-1].pan = Integer.parseInt((String)sel_status.fields.elementAt(0));
 		  	camSim[id-1].tilt = Integer.parseInt((String)sel_status.fields.elementAt(1));
 		  	camSim[id-1].zoom = Integer.parseInt((String)sel_status.fields.elementAt(2));
 		  	
 		  	logFileWriter.println("Current physical status of Camera " + id+ ": pan = " + camSim[id-1].pan + "  tilt = " + camSim[id-1].tilt + "  zoom = " + camSim[id-1].zoom + ".");
            logFileWriter.println();
            		  	
 		  	startTime = System.currentTimeMillis();
 		  	requests.removeElement((ActionRequest)sel_ar); 		  	
 		  	updateCosts();
 		  	endTime = System.currentTimeMillis();
 		  	overallSchedulingTime += endTime - startTime;		  		     	
 		  } 
 	 		  
 	    logFileWriter.println("#");
 	    for (i=0;i<didVector.size();i++)
 	      {
 	      	logFileWriter.println("Service time on Device " + (i+1) + ": " + singleServiceTime[i] + " ms.");
 	      	if (singleServiceTime[i] > overallServiceTime)
 	      	  overallServiceTime = singleServiceTime[i]; 
 	      	if (i == singleServiceTime.length-1)
 	      	  logFileWriter.println("#");
 	      	logFileWriter.println();		      	
 	      }
 		
 		logFileWriter.println("#");
 		logFileWriter.println("Service time: " + overallServiceTime + " ms.");
 		logFileWriter.println("Scheduling time: " + overallSchedulingTime + " ms.");
 		logFileWriter.println("Makespan: " + (overallServiceTime + overallSchedulingTime) + " ms.");
 		logFileWriter.println("#");
 	  } 	
 	
 	
 	/** Select the current shorest request on all canidate cameras. */
 	public void shorestRequestSelection() 
 	  {
		int i,j;
		double minActionCost = Double.MAX_VALUE;
		
		for (i=0;i<requests.size();i++)
		  for (j=0;j<((ActionRequest)requests.elementAt(i)).costVector.size();j++)
		    if ( minActionCost > Double.parseDouble((String)((ActionRequest)requests.elementAt(i)).costVector.elementAt(j)) )
		      {
		      	sel_ar = (ActionRequest)requests.elementAt(i);
		      	minActionCost = Double.parseDouble((String)sel_ar.costVector.elementAt(j));
		      	sel_did = (String)sel_ar.didVector.elementAt(j);
		      	sel_status = (Tuple)sel_ar.statusVector.elementAt(j);
		      	total_actionCost = minActionCost;			      		      	
		      }
      }
	
	
	/** Update the cost of other requests on the camera that services the currently selected request. */
	public void updateCosts() 
	  {
		int i, j;
		int pan, tilt, zoom;
		int deltaPan, deltaTilt, deltaZoom;
		double headMovementCost, actionCost;
		Tuple t;
		ActionRequest ar;		
		
		for (i=0;i<requests.size();i++)
		  {
		  	ar = (ActionRequest)requests.elementAt(i);
		  	for (j=0;j<ar.didVector.size();j++)
		      if (((String)ar.didVector.elementAt(j)).equals(sel_did))
		       {
		         t = (Tuple)ar.statusVector.elementAt(j);
		         pan = Integer.parseInt((String)t.fields.elementAt(0));
    	   	     tilt = Integer.parseInt((String)t.fields.elementAt(1));
    	   	     zoom = Integer.parseInt((String)t.fields.elementAt(2));
    	   	     deltaPan = Math.abs(camSim[Integer.parseInt(sel_did)-1].pan - pan);
    	   	     deltaTilt = Math.abs(camSim[Integer.parseInt(sel_did)-1].tilt - tilt);
    	   	     deltaZoom = Math.abs(camSim[Integer.parseInt(sel_did)-1].zoom - zoom);
    	   	     headMovementCost = camSim[Integer.parseInt(sel_did)-1].computeHeadMovementCost(deltaPan, deltaTilt, deltaZoom);
		         actionCost = estimateActionCost(headMovementCost);     	   	  	
		         actionCost += total_actionCost;
		         
		         ar.costVector.removeElementAt(j);
		         ar.costVector.insertElementAt(String.valueOf(actionCost),j);
		       } 
		   }	
	  }
	  
	  
	/** Estimate the cost of a <code>photo()</code> action. */
	public double estimateActionCost (double headMovementCost)
      {
      	return 2*CameraSimulator.STANDARD_CONNECTION_TIME + headMovementCost + CameraSimulator.TAKE_MEDIUM_SIZE_PHOTO_ESTIMATED_COST;
      }   	 
 	
 } 