/*

* Copyright (c) 2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

import java.util.*;

/**
 * Generate a synthetic action workload with skewed distribution.
 *
 * @see AortaSchedulingSimulator
 * @see UniformSyntheticWorkload
 */
public class SkewedSyntheticWorkload {
	
    /** Number of cameras involved in the scheduling. */
    public int cameraNum;
    
    /** Number of action requests involved in the scheduling. */
    public int requestNum;
     
    /** A value in [0,1] that indicates the portion of the total number of cameras with a skewed workload on. */
    public double skewness; 
    
    Random ran = new Random(); 
    
                                       
	public SkewedSyntheticWorkload (int cameraNum, int requestNum, double skewness) 
	  {
		this.cameraNum = cameraNum;
		
		this.requestNum = requestNum;
		
		this.skewness = skewness;
	  }
	
	
	/** 
	 * Generate the skewed workload. 
	 * Half of the requests are skewed, half are not. 
	 */
	public Vector generate () 
	  {
		int i;
		Vector requests = new Vector (30,10);		
		
		for (i=0;i<(requestNum/2);i++)
		  generateUniformRequest(requests, i+1); 
		  		
		for (;i<requestNum;i++)
		  generateSkewedRequest(requests, i+1); 		
		
		return requests;		
	  } 
	  
	  
	/** Generate a single skewed action request. */  
	public void generateSkewedRequest(Vector requests, int id) 
	  {
		ActionRequest ar = new ActionRequest ();
		
		ar.aid = String.valueOf(id);			
		
	    double cost;
		int skewedCameraNum = (int)(cameraNum * skewness);		
		for (int i=0;i<skewedCameraNum;i++)
		  {
		   	ar.didVector.addElement(String.valueOf(i+1));
		  	
		  	cost = ranActionCost(ar);
		  	ar.costVector.addElement(String.valueOf(cost));	  	
		  } 
		
		requests.addElement((ActionRequest)ar);			
	  } 
	 
		
	/** Generate a single uniform action request. */
	public void generateUniformRequest(Vector requests, int id)  
	  {
		ActionRequest ar = new ActionRequest ();
		
		ar.aid = String.valueOf(id);						
		
	    double cost;
		for (int i=0;i<cameraNum;i++)
		  {
		  	ar.didVector.addElement(String.valueOf(i+1));
		  	
		  	cost = ranActionCost(ar);
		  	ar.costVector.addElement(String.valueOf(cost));	  	
		  } 
		
		requests.addElement((ActionRequest)ar);		
      } 
       
    
    /** Randomly generate the cost a <code>photo()</code> action. */
    public double ranActionCost (ActionRequest ar)
     {
       CameraSimulator camSim = new CameraSimulator("1");
       
       int pan = ranPan(); 
       int tilt = ranTilt();
       int zoom = ranZoom(); 
       
       Vector v = new Vector(10,5);
       v.addElement(String.valueOf(pan));
       v.addElement(String.valueOf(tilt));
       v.addElement(String.valueOf(zoom));
       Tuple t = new Tuple(v);
       ar.statusVector.addElement((Tuple)t);
       
       double headMovementCost = camSim.computeHeadMovementCost(pan, tilt, zoom);
       
       return 2*camSim.STANDARD_CONNECTION_TIME + headMovementCost + camSim.TAKE_MEDIUM_SIZE_PHOTO_ESTIMATED_COST; 
     }
     
    /** Randomly pick a target <code>pan</code> for the camera head. */
    public int ranPan()
	  {
	  	return CameraSimulator.MIN_PAN + Math.abs(ran.nextInt()) % (CameraSimulator.MAX_PAN - CameraSimulator.MIN_PAN + 1);
	  }
	  
    /** Randomly pick a target <code>tilt</code> for the camera head. */
    public int ranTilt()
	  {
	  	return CameraSimulator.MIN_TILT + Math.abs(ran.nextInt()) % (CameraSimulator.MAX_TILT - CameraSimulator.MIN_TILT + 1);
	  }
	  
    /** Randomly pick a target <code>zoom</code> for the camera head. */
    public int ranZoom()
	  {
	  	int num = Math.abs(ran.nextInt()) % 4;
	  	
	  	if (num > 1) 
	  	  num = Math.abs(ran.nextInt()) % 4;
	  	
	  	if (num > 2)
	  	  num = Math.abs(ran.nextInt()) % 4;
	  	
	  	if (num == 0) return 1;
	  	else if (num == 1) return 2501; 
	         else if (num == 2) return 5001;
	              else return 9999;	  	
	  }	  
	
} 